// GPIODBC.cpp : Defines the entry points for the DLL application.
//

#include "stdafx.h"
#include "./resource.h"
#include "./querywizard.h"
#include "./shapepointcreator.h"

void Load(SUHANDLE hWorkspace);
bool LoadInternal(SUHANDLE hWorkspace, CRecordset *prs, short nXCol, short nYCol, SUHANDLE *hThemeOut);
SURESULT LoadFromJS(SUHANDLE hWorkspace, SUJavaScriptInvokeHookData *args);

//--------------------------------------------------------
// Plug-in local data
//
// Every plug-in needs an instance of SUExport called suapi
SUExport suapi;

//--------------------------------------------------------
// Forward declarations
//
SURESULT __cdecl HookFunc(SUHANDLE hWorkspace, int hookType, SUHookData *args);

//--------------------------------------------------------
// fnVersion()
//    Tells ShapeUp what version of the API we are aware
//    of.
//
size_t __cdecl fnVersion()
{
	return SHAPEUP_API_VERSION;
}


//--------------------------------------------------------
// fnGUID()
//    Identifies this plug-in.
//
const GUID* __cdecl fnGUID()
{
	// {18999D7C-B47C-41B0-862F-15A3A6E138EB}
	static const GUID guid = 
		{ 0x18999D7C, 0xB47C, 0x41B0, { 0x86, 0x2F, 0x15, 0xA3, 0xA6, 0xE1, 0x38, 0xEB } };
	return &guid;
}


//--------------------------------------------------------
// fnInit()
//    This function gets called just after the plug-in is
//    loaded in the ShapeUp start sequence.
//
SURESULT __cdecl fnInit(SUExport *pExport)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	// Tell ShapeUp who this plug-in is, and what it wants

	// Set some texts, which will be displayed here and there in ShapeUp
	pExport->import.pszName = "GPIODBC";
	pExport->import.pszDescription = "This plug-in reads point data from a database table.\nODBC point loader.";
	pExport->import.pszInfo =
		"This plug-in reads point data from a database table.\n\n"
		"It is also JavaScript enabled, and the following methods are available:\n"
		"  loadLayer(connectionString, queryString, eastcol, northcol) loads point data into the current workspace. The method returns the layer if successful.\n"
		"    connectionString, an ODBC connection string.\n"
		"    queryString, an SQL select statement.\n"
		"    eastcol and northcol are zero-based indecies indicating which columns should be used as coordinates.\n"
		"  usage:\n"
		"    var dbGUID = '{18999D7C-B47C-41B0-862F-15A3A6E138EB}';\n"
		"    var dbloader = new Plugin(dbGUID);\n"
		"    var layer = dbloader.loadLayer('DSN=MYDB;UID=guest;', 'Select lat, long, others From wgsdata', 1, 0);\n"
		"\nCopyright (c) Niklas Lindquist 2008";
	pExport->import.pszPluginVersion = "1.00";

	// Now keep the entire struct which is the actual ShapeUp API
	// (we need the function pointers later)
	::memcpy(&suapi, pExport, sizeof(SUExport));

	// Set-up our hooks
	pExport->import.pfnHook = HookFunc;

	// Set up our hooks here using the ShapeUp_AddXxxHook() functions

	// We want an entry in the Insert | Add Theme From sub menu:
	ShapeUp_AddLoadLayerHook(0, "ODBC Query...", "Load data from an SQL query", NULL, NULL);

	// We also want to be JavaScript enabled:
	const char* properties[] =
	{
		NULL
	};
	const char* methods[] =
	{
		"loadLayer",
		NULL
	};
	ShapeUp_AddJavaScriptInvokeHook(0, properties, methods);

	return SURESULT_OK;
}


//--------------------------------------------------------
// fnTerminate()
//    This function gets called just before this plug-in
//    is unloaded by ShapeUp. Clean up whatever needs to
//    be cleaned up.
//
void __cdecl fnTerminate()
{
	// TODO: Add clean-up code if needed
}


//--------------------------------------------------------
// HookFunc()
//    This is our main function getting notified of state
//    changes in ShapeUp. We check the type of change, and
//    delegate.
//
SURESULT __cdecl HookFunc(SUHANDLE hWorkspace, int hookType, SUHookData *args)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	switch (hookType)
	{
	case HOOKTYPE_MENUITEM:
		Load(hWorkspace);
		break;
	case HOOKTYPE_JAVASCRIPTINVOKE:
		return LoadFromJS(hWorkspace, (SUJavaScriptInvokeHookData*)args);
	default:
		break;
	}

	return SURESULT_OK;
}


//--------------------------------------------------------
// Load()
//    This function opens the data definition dialog.
//
void Load(SUHANDLE hWorkspace)
{
	CQueryWizard dlg("ODBC data selector");
	if (dlg.DoModal() == ID_WIZFINISH)
	{
		CWaitCursor wc;
		SUWorkspace ws(hWorkspace);
		SUStatusBar sb;
		sb.SetText("Loading objects...");
		LoadInternal(hWorkspace, dlg.m_pRS, (short) dlg.m_nXCol, (short) dlg.m_nYCol, NULL);
		sb.SetText(NULL);
	}
}


//--------------------------------------------------------
// LoadFromJS()
//    This function is called when a JavaScript called us.
//
SURESULT LoadFromJS(SUHANDLE hWorkspace, SUJavaScriptInvokeHookData *args)
{
	if (0 == strcmp("loadLayer", args->name))
	{
		// We need 4 parameters to proceed
		if (args->argc < 4)
			return SURESULT_E_ILLEGALPARAM;

		// Check parameter types
		if (args->args[0].type != DT_STRING ||
			args->args[1].type != DT_STRING ||
			args->args[2].type != DT_INT ||
			args->args[3].type != DT_INT)
			return SURESULT_E_ILLEGALPARAM;

		// Set a default return value of null (in case something fails)
		args->retc = 1;
		ShapeUp_Allocate(sizeof(SUJavaScriptArgument) * args->retc, (void**)&args->rets);
		args->rets[0].type = DT_SUHANDLE;
		args->rets[0].handleVal = NULL;

		args->flags = JSIHDF_TREATRETURNASSINGLEVALUE;

		CString strConnectionString, strQuery;
		int nXCol = 0, nYCol = 1;

		// Get data
		strConnectionString = args->args[0].stringVal;
		strQuery = args->args[1].stringVal;
		nXCol = args->args[2].intVal;
		nYCol = args->args[3].intVal;

		// Query sanity check. No 'drop database' on our watch.
		// Maybe we should allow a CALL also?
		CString q = strQuery;
		q = q.TrimLeft();
		if (q.Left(6).CompareNoCase(_T("SELECT")) != 0)
			return SURESULT_E_ILLEGALPARAM;

		// Connect
		try
		{
			CDatabase db;
			if (db.OpenEx(strConnectionString, CDatabase::noOdbcDialog | CDatabase::openReadOnly))
			{
				CRecordset rs(&db);
				if (rs.Open(CRecordset::snapshot, strQuery))
				{
					if (!rs.IsBOF())
					{
						SUHANDLE hTheme = NULL;
						if (LoadInternal(hWorkspace, &rs, nXCol, nYCol, &hTheme))
						{
							// Now hTheme contains our return value
							args->rets[0].handleVal = hTheme;
							return SURESULT_OK;
						}
					}
				}
			}
		}
		catch (CDBException *ex)
		{
			ex->ReportError();
			ex->Delete();
		}
		catch (CMemoryException *ex)
		{
			ex->ReportError();
			ex->Delete();
		}
	}

	return SURESULT_OK;
}


//--------------------------------------------------------
// LoadInternal()
//    Load the geo info from the database.
//
bool LoadInternal(SUHANDLE hWorkspace, CRecordset *prs, short nXCol, short nYCol, SUHANDLE *hThemeOut)
{
	if (!prs)
		return false;

	CShapePointCreator creator(0);
	BYTE *pShapeBytes = new BYTE[creator.GetAllocSizeBytes()];
	const char **attrs = NULL;
	
	SUWorkspace ws(hWorkspace);
	SUTheme thm;

	try
	{
		prs->MoveFirst();

		thm = ws.CreateNewTheme(Shape_Point, "ODBC Points");

		thm.BatchAddShapeBegin();

		CString str;
		CODBCFieldInfo info;
		short nFieldCount = prs->GetODBCFieldCount();

		// Set-up headers
		for (short f = 0; f < nFieldCount; ++f)
		{
			prs->GetODBCFieldInfo(f, info);
			unsigned long  ulType = ATTRIBUTETYPE_STRING;
			switch (info.m_nSQLType)
			{
			case SQL_BIT:
				ulType = ATTRIBUTETYPE_BOOL;
				break;
			case SQL_REAL:
			case SQL_FLOAT:
			case SQL_DOUBLE:
				ulType = ATTRIBUTETYPE_DOUBLE;
				break;
			case SQL_C_LONG:
			case SQL_C_SHORT:
			case SQL_C_TINYINT:
			case SQL_C_SLONG:
			case SQL_C_SSHORT:
			case SQL_C_STINYINT:
			case SQL_C_ULONG:
			case SQL_C_USHORT:
			case SQL_C_UTINYINT:
				ulType = ATTRIBUTETYPE_INT;
				break;
			default: break;
			}
			thm.AddAttribute(ulType, info.m_strName, NULL);
		}

		CStringArray arr;
		attrs = new const char*[nFieldCount];

		// Add data
		for (int i = 0; !prs->IsEOF(); ++i)
		{
			for (short f = 0; f < nFieldCount; ++f)
			{
				prs->GetFieldValue(f, str);
				arr.Add(str);
				attrs[f] = arr[f];
			}
			double x = atof(arr[nXCol]);
			double y = atof(arr[nYCol]);
			creator.Write(pShapeBytes, x, y);
			SUShape shp = thm.AddShape((CShapeRecord*) pShapeBytes);
			shp.SetAttributes(thm, nFieldCount, attrs);
			arr.RemoveAll();
			prs->MoveNext();
		}

		thm.BatchAddShapeEnd();
		if (ws.GetThemeCount() == 1)
		{
			ws.GetMapWindow().ZoomToExtent(thm);
			thm.Select(true);
		}
		ws.Refresh();

		delete [] attrs;
		delete [] pShapeBytes;

		if (hThemeOut)
			*hThemeOut = thm;

		return true;
	}
	catch (CDBException *ex)
	{
		if (thm.IsValid())
			thm.BatchAddShapeEnd();
		try { ws.RemoveTheme(thm); } catch (...) {}
		ex->ReportError();
		ex->Delete();
	}
	catch (CMemoryException *ex)
	{
		if (thm.IsValid())
			thm.BatchAddShapeEnd();
		try { ws.RemoveTheme(thm); } catch (...) {}
		ex->ReportError();
		ex->Delete();
	}
	catch (SUException ex)
	{
		AfxMessageBox(ex.GetErrorDesc().c_str(), MB_OK | MB_ICONERROR);
		if (thm.IsValid())
			thm.BatchAddShapeEnd();
		try { ws.RemoveTheme(thm); } catch (...) {}
	}

	delete [] attrs;
	delete [] pShapeBytes;
	return false;
}


